@@ -3,8 +3,6 @@
  *   based on the Intel CRC32 instruction
  *   provided in the Intel SSE4.2 instruction set
  *
- *   If SSE4.2 is not supported on the platform, this module will not be able to get installed
- *
  *    ICRAR - International Centre for Radio Astronomy Research
  *    (c) UWA - The University of Western Australia, 2014
  *    Copyright by UWA (in the framework of the ICRAR)
@@ -37,7 +35,7 @@
 static crc_function crc_fn;
 
 static
-PyObject* crc32c_crc32(PyObject *self, PyObject *args) {
+PyObject* crc32c_crc32c(PyObject *self, PyObject *args) {
 	Py_buffer pbin;
 	unsigned char *bin_data = NULL;
 	uint32_t crc = 0U, result;
@@ -63,25 +61,62 @@ PyObject* crc32c_crc32(PyObject *self, PyObject *args) {
 	return PyLong_FromUnsignedLong(result);
 }
 
+static
+PyObject *crc32c_crc32(PyObject *self, PyObject *args)
+{
+	if (PyErr_WarnEx(PyExc_DeprecationWarning,
+	                 "crc32c.crc32 will be eventually removed, use crc32c.crc32c instead",
+	                 1) == -1) {
+		return NULL;
+	}
+	return crc32c_crc32c(self, args);
+}
+
+/* The different values the SW mode preference can take */
+enum crc32c_sw_mode {
+	UNSPECIFIED,
+	AUTO,
+	FORCE,
+	NONE
+};
+
+static enum crc32c_sw_mode get_sw_mode(void)
+{
+	char *sw_mode = getenv("CRC32C_SW_MODE");
+	if (sw_mode == NULL) {
+		return UNSPECIFIED;
+	}
+	else if (!strcmp(sw_mode, "auto")) {
+		return AUTO;
+	}
+	else if (!strcmp(sw_mode, "force")) {
+		return FORCE;
+	}
+	else if (!strcmp(sw_mode, "none")) {
+		return NONE;
+	}
+	return UNSPECIFIED;
+}
 
 static PyMethodDef CRC32CMethods[] = {
-	{"crc32",   crc32c_crc32,   METH_VARARGS, "Calculate crc32c using Intel SSE4.2 instruction."},
+	{"crc32",   crc32c_crc32,   METH_VARARGS, "Calculate crc32c incrementally (deprecated)"},
+	{"crc32c",  crc32c_crc32c,  METH_VARARGS, "Calculate crc32c incrementally"},
 	{NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
 static const char *import_error_msg = "\n\n"
-"SSE4.2 extensions providing a crc32c hardware instruction are not available in\n"
-"your processor. If you still need to use the crc32c checksum algorithm this\n"
-"package comes with a software implementation that can be loaded instead. For\n"
-"that set the CRC32C_SW_MODE environment variable before loading the package to\n"
-"one of the following values:\n\n"
-" * 'auto' to use software implementation if no CPU hardware support is found.\n"
-" * 'force' to use software implementation regardless of CPU hardware support.\n"
-" * '1' means 'force', but will be eventually discontinued.\n";
+"Hardware extensions providing a crc32c hardware instruction are not available in\n"
+"your processor. This package comes with a software implementation, but this\n"
+"support has been opted out because the CRC32C_SW_MODE environment variable is\n"
+"set to \"none\". CRC32C_SW_MODE can take one of the following values:\n"
+" * If unset: use the software implementation if no hardware support is found\n"
+" * 'auto': as above, but will eventually be discontinued\n"
+" * 'force': use software implementation regardless of hardware support.\n"
+" * 'none': fail if no hardware support is found (this error).\n";
 
 /* Support for Python 2/3 */
 #if PY_MAJOR_VERSION >= 3
-static struct PyModuleDef moduledef = {PyModuleDef_HEAD_INIT, "crc32c", "wrapper for crc32c Intel instruction", -1, CRC32CMethods};
+static struct PyModuleDef moduledef = {PyModuleDef_HEAD_INIT, "crc32c", "crc32c implementation in hardware and software", -1, CRC32CMethods};
 	#define MOD_INIT(name) PyMODINIT_FUNC PyInit_##name(void)
 	#define MOD_DEF(m, name, doc, methods) \
 		m = PyModule_Create(&moduledef);
@@ -96,33 +131,43 @@ static struct PyModuleDef moduledef = {PyModuleDef_HEAD_INIT, "crc32c", "wrapper
 MOD_INIT(crc32c)
 {
 	PyObject *m;
+	PyObject *hardware_based;
+	enum crc32c_sw_mode sw_mode;
 
-	char *sw_mode = getenv("CRC32C_SW_MODE");
-	int force_sw_impl = sw_mode != NULL && (!strcmp(sw_mode, "1") || !strcmp(sw_mode, "force"));
-	int fallback_to_sw_impl = sw_mode != NULL && !strcmp(sw_mode, "auto");
-
+	sw_mode = get_sw_mode();
 	crc_fn = NULL;
-	if (force_sw_impl) {
+	if (sw_mode == FORCE) {
 		crc_fn = _crc32c_sw_slicing_by_8;
+		hardware_based = Py_False;
 	}
 #if defined(IS_INTEL)
 	else if (_crc32c_intel_probe()) {
 		crc_fn = _crc32c_hw_adler;
 		crc32c_init_hw_adler();
+		hardware_based = Py_True;
+	}
+#elif defined(IS_ARM) && (defined(__linux__) || defined(linux))
+	else if (_crc32c_arm64_probe()) {
+		crc_fn = _crc32c_hw_arm64;
+		hardware_based = Py_True;
 	}
 #endif
-	else if (fallback_to_sw_impl) {
+	else if (sw_mode == UNSPECIFIED || sw_mode == AUTO) {
 		crc_fn = _crc32c_sw_slicing_by_8;
+		hardware_based = Py_False;
 	}
-
-	else {
+	else if (sw_mode == NONE) {
 		PyErr_SetString(PyExc_ImportError, import_error_msg);
 		return MOD_VAL(NULL);
 	}
 
-	MOD_DEF(m, "crc32c", "wrapper for crc32c Intel instruction", CRC32CMethods);
+	MOD_DEF(m, "crc32c", "crc32c implementation in hardware and software", CRC32CMethods);
 	if (m == NULL) {
 		return MOD_VAL(NULL);
 	}
+	Py_INCREF(hardware_based);
+	if (PyModule_AddObject(m, "hardware_based", hardware_based) < 0) {
+		return MOD_VAL(NULL);
+	}
 	return MOD_VAL(m);
 }
